//---------------------------------------------------------------------------
// File  :  \CBproj\SoundUtl\NoiseBlanker.cpp
// Date  :  2004-04-29  (YYYY-MM-DD)
// Author:  Wolfgang Buescher  (DL4YHF)
//
// Description:
//     Interface for the CNoiseBlanker class.
//
// Revision history :
//     2002-08-24   Created
//     2004-03-29   Almost completely rewritten; now without FFT
//                   (simpler but more robust now)
//     2010-05-22   Cleaned up some old mess; 
//                   temporarily uploaded the source to the SL website .
//
//---------------------------------------------------------------------------

#include "SWITCHES.H"  // project specific compiler switches ("options")
                       // must be included before anything else !



#include <windows.h>
#include <math.h>

#pragma hdrstop   // no precompiled headers after this point
#pragma warn -8004 // <var is a assigned a value that is never used> - so what

#include "utility1.h"    // uses UTL_NamedMalloc instead of malloc (for debugging)
#include "SoundTab.h"    // frequently used types and tables for audio processing
#include "NoiseBlanker.h"


/*** From DF6NM's mail ..

     Title:   LF: Noise blanking for narrowband modes?
     Date:    2002-08-12

Narrowband reception goes along with long FFT integration times, and in a 
linear system, the effective noise level is given by the total energy of
background and burst noise. If for example the QRN covers 10% of the
aquisition time with a level of 30 dB above the background, it would look 
like 100-fold increased Gaussian noise, a desastrous 20 dB SNR loss. If we 
however clip the QRN at say 6 dB above background, the average noise energy
will be increased by only 40% (1.5 dB). And the signal will virtually 
disappear during 10% of the time (-0.8dB), adding up to only 2.3 dB SNR 
degradation.

But I am wondering, if not a noise-blanking scheme would be more effective 
than simple limiting. In a software implementation with some buffering, one
could even avoid switching clicks by smoothly reducing the gain around each
burst (eg. a 10 ms cosine shape), keeping the spectral widening of strong
carriers moderate. The user could preset a fraction of blanked time (e.g 10%, 
dependent on the severeness of the QRN), and the SW could then automatically 
control the threshold to approach that.

Going one step further, we could try to give an optimal estimate of a signal
in the presence of a time-variing noise level. In analogy to a Wiener filter,
the noise contribution of each sample (ie. 10 ms block) to the FFT input
would then have to be inversly proportional to its SNR. If the noise level
went up temporarily by 3 dB, we should reduce the gain by 6 dB. For high
noise peaks, this will of course result in almost complete blanking. One nice
feature of the scheme is that we don't have to worry about a threshold, all
we have to do is keep track of the wideband power from the RX (assuming that
this is noise and our weak signal is below that), and dynamically control the
gain.

     By DL4YHF:
In an initial trial from 2002-05, an FFT was used.
   To detect the noise level, a short real-valued FFT and G4JNT's "CFAR"
   noise detection method may be used.

Later (2004-03) this was discarded and replaced with a more conventional
   noiseblanker, "emulating" the function in many shortwave receivers.
   The principle is now quite similar to the 2nd paragraph in DF6NM's mail.
   A buffer capable of storing about 50 milliseconds of the 'latest'
   samples is used as INPUT DELAY LINE and -at the same time- to calculate
   a moving average for the "QRN energy" in the last 50 milliseconds.
   This parameter controls the gain factor, which is applied to the
   delayed samples (so the smooth gain reduction starts before the QRN pulse
   actually "takes effect" ).

***/



//----------------------- Constants ---------------------------------------




//----------------------- Data Types --------------------------------------



//----------------------- Helper Routines ---------------------------------


//---------------------------------------------------------------------
void FloatQuickSort( float *A, int AHigh, int iLo, int iHi )
{
  // Quick Sort Algorithm for increasing order.
  int    lo, hi;
  T_Float Mid;
  float  t_swap;

  lo = iLo;
  hi = iHi;
  Mid = A[(lo+hi)/2];

  do
   {
    while (A[lo] < Mid)
        lo++;
    while (A[hi] > Mid)
        hi--;
    if (lo <= hi)
    {
      t_swap = A[lo];
      A[lo] = A[hi];
      A[hi] = t_swap;
      lo++;
      hi--;
    }
   }
  while (lo <= hi);

  if (hi > iLo)
    FloatQuickSort(A, AHigh, iLo, hi);
  if (lo < iHi)
    FloatQuickSort(A, AHigh, lo, iHi);
}





//***************************************************************************
//  Implementation of the  CNoiseBlanker   class
//***************************************************************************


//---------------------------------------------------------------------------
CNoiseBlanker::CNoiseBlanker()     // constructor
{
 int i;
#if(NOISEBLANKER_USE_FFT)  // NOISEBLANKER_USE_FFT usually defined as 0 in *.h
   m_pFft = NULL;
   m_wFftInputIndex    = 0;
#else  // no FFT but a quite conventional design, with a delay line...
#endif // NOISEBLANKER_USE_FFT

   m_fltAvrgNoisePower = -100;
   m_fltCurrNoiseLevel = -100;
   m_fltPeakNoisePower = -100;
   m_fltCurrBlankingMultiplier = 1.0;

   m_fltMinPeak2AvrgPowerRatio = 20; // peak must exceed average POWER by X times

   m_dwBufferSize = 0;
   m_pFltBuffer = NULL;
   m_dwBufIndex = 0;

   m_iRampIndex = m_iRampLength = m_iRampDir = 0;
   for(i=0; i<NBL_AVRG_LP_STAGES; ++i)
     m_fltAvLowpass[i]=0.0;

} // end CNoiseBlanker::CNoiseBlanker()



//---------------------------------------------------------------------------
CNoiseBlanker::~CNoiseBlanker() // destructor, clean up
{
  CNoiseBlanker::Exit();
} // end CNoiseBlanker::~CNoiseBlanker() [destructor]


//---------------------------------------------------------------------------
BOOL CNoiseBlanker::Init(  int iMode, T_Float fltSampleRate )
   // Caution: Do not re-initialized the noiseblanker
   //          while ProcessSamples() is called in another thread !
{
  m_fltSampleRate  = fltSampleRate;
  if(m_fltSampleRate<1)
     m_fltSampleRate=0;   // crude way to avoid div by zero later

#if( NOISEBLANKER_USE_FFT )
  m_wFftInputIndex = 0;
  if(m_pFft)
   { delete m_pFft;
     m_pFft = NULL;
   }
  m_pFft = new Cfft(
           NBL_FFT_SIZE,    // int fft_size,
           FALSE,  // bool use_complex_input, FALSE=real values for INPUT, TRUE=complex
           FALSE,  // bool need_phase_output, FALSE=amplitude only, TRUE=also need phase info
           FFT_WINDOW_HANN, // FFT window function
           0 );    // number of AVERAGE channels (zero, one or two)
#else  // no FFT ... but a delay line :


#endif // NOISEBLANKER_USE_FFT

  m_dwBufIndex = 0;
  m_iRampIndex = 0;
  m_fltOldRampTime = -1.0; // to force buffer allocation..
  m_fltRampTime = 0.01;
  m_dwBufferSize = 0;

  return TRUE;
} // end CNoiseBlanker::Init( )

//---------------------------------------------------------------------------
void CNoiseBlanker::Exit( void )
   // Frees all resources without 'destruction' .
   // (allows to do this when THE APPLICATION wants to, not dictated by the runtime lib)
{
#if(NOISEBLANKER_USE_FFT)
  if(m_pFft)
   { delete m_pFft;
     m_pFft = NULL;
   }
#else  // no FFT ... but a delay line :

#endif // NOISEBLANKER_USE_FFT

  if( m_pFltBuffer )  // does a circular buffer exist ?
   { UTL_free(m_pFltBuffer);
     m_pFltBuffer = NULL;
   }
} // end CNoiseBlanker::Exit()


//---------------------------------------------------------------------------
T_Float CNoiseBlanker::GetRampTime(void)
{ return m_fltRampTime;
}
BOOL CNoiseBlanker::SetRampTime(T_Float fltRampTime)
{ m_fltRampTime = fltRampTime;
  return TRUE;
}
void CNoiseBlanker::ApplyNewRampTime(void)
{
 DWORD dwNewBufferSize;
 BOOL  fResult=TRUE;

  // The buffer must be -at least- large enough for the 'ramp time' .
  // The same buffer is also used to keep the sum of squares up-to-date .
  m_iRampLength = m_fltRampTime * m_fltSampleRate;
  if(m_iRampLength<1)
     m_iRampLength=1;
  if(m_iRampLength>32767)  // note: 'int' is 32 bit these days.. at least !
     m_iRampLength=32767;

  dwNewBufferSize = m_iRampLength;
  // Emergency limits for the allocated buffer size :
  //   ( too large buffer lets the CPU run out of cache memory,
  //     result = severely reduced performance )
  if( dwNewBufferSize < 10 )
      dwNewBufferSize = 10;
  if( dwNewBufferSize > 65536 )  // enough for more than a second at 48kHz
      dwNewBufferSize = 65536;
  // Allocate the buffer if it doesn't already exist in the same size :
  if( dwNewBufferSize != m_dwBufferSize )
   { if( m_pFltBuffer )
      { UTL_free(m_pFltBuffer); m_pFltBuffer=NULL;
        m_dwBufferSize = 0;
      }
   }
  if( ! m_pFltBuffer )   // allocate a new buffer ?
   { m_pFltBuffer = (T_Float*)UTL_NamedMalloc( "NoiseBlk", dwNewBufferSize * sizeof(T_Float) );
     if( m_pFltBuffer )
      {
        m_dwBufferSize = dwNewBufferSize;
        // To fill the gap of 'emitted' samples, because the caller of
        // ProcessSamples() wants to retrieve the same count of samples
        // as he passed in, the buffer MUST be initialized with valid data:
        for(DWORD dwTemp=0; dwTemp<m_dwBufferSize; ++dwTemp)
            m_pFltBuffer[dwTemp] = 0.0;
      }
     else // (UTL_-)malloc failed :
      {
        fResult=FALSE;
      }
     m_dwBufIndex = 0;
     m_iRampIndex = 0;
   } // end if < allocate new buffer ? >

  m_fltOldRampTime = m_fltRampTime;

} // end CNoiseBlanker::SetRampTime()


//---------------------------------------------------------------------------
BOOL CNoiseBlanker::SetTriggerLevel( T_Float fltMinPeakToAvrg_dB )
{
  // Convert the DECIBEL VALUE into a POWER RATIO as required for the detector:
  m_fltMinPeak2AvrgPowerRatio = pow(10.0, fltMinPeakToAvrg_dB / 10.0);

  return TRUE;
} // end CNoiseBlanker::SetTriggerLevel()

//---------------------------------------------------------------------------
BOOL CNoiseBlanker::SetMaxPulseWidth( T_Float fltNoisePulseWidth_sec )
{
  fltNoisePulseWidth_sec = fltNoisePulseWidth_sec;  // not implemented yet !
  return FALSE;
}

//---------------------------------------------------------------------------
BOOL CNoiseBlanker::SetMaxPulseRiseTime( T_Float fltNoiseRiseTime_sec )
{
  fltNoiseRiseTime_sec = fltNoiseRiseTime_sec;      // not implemented yet !
  return FALSE;  
}

//---------------------------------------------------------------------------
BOOL CNoiseBlanker::ProcessSamples(
               T_Float *pSource, T_Float *pDest,
               int iNrSamplePoints )
{

  int iSampleIndex;
  T_Float fltTemp;
  T_Float fltNoisePower;

  // Correct a few parameters once before entering the loop..
  if(m_fltOldRampTime != m_fltRampTime)
   { ApplyNewRampTime();
     m_fltOldRampTime = m_fltRampTime;
   }
  if(m_dwBufIndex>= m_dwBufferSize)
     m_dwBufIndex = 0;
  if(m_iRampIndex>=m_iRampLength)
     m_iRampIndex= m_iRampLength-1;

  fltNoisePower = m_fltAvrgNoisePower;  // initial value from previous chunk

  for( iSampleIndex=0; iSampleIndex<iNrSamplePoints; ++iSampleIndex)
   {
     fltTemp = *pSource++;

     // enter new sample in buffer ...
     m_pFltBuffer[m_dwBufIndex++] = fltTemp;
     if(m_dwBufIndex>=m_dwBufferSize)
        m_dwBufIndex = 0;


     // Detect "pulse-noise level" at the input .
     //  (was once utterly complicated with frequency-selective detection,
     //   but later greatly simplified with a non-frequency-selective
     //   noise level detection; see below)
#if( NOISEBLANKER_USE_FFT )  // not in use !
     if( m_wFftInputIndex < NBL_FFT_SIZE)
         m_fltFftInput[m_wFftInputIndex++] = fltTemp;
     if(  (m_pFft) &&  (m_wFftInputIndex >= NBL_FFT_SIZE) )
      {  // time to calculate the next FFT ..
         m_wFftInputIndex = 0;
         m_pFft->CalcFFT( m_fltFftInput,
                       0, // int iAve,
                       0, // int iAvrgChannel, 0=1st average channel, 1=2nd average channel
                   FALSE, // TRUE=logarithmic output, FALSE=linear
          m_fltFftOutput, // float* pfltOutBuf,  destination buffer (here: amplitudes)
           NBL_FFT_SIZE); // max count of entries in pfltOutBuf for safety

         // Sort the array into order of increasing amplitude.
         FloatQuickSort( m_fltFftOutput, NBL_FFT_SIZE/2-1, 0, NBL_FFT_SIZE/2-1);

         // The amplitude of the lower quartile value (eg bin number 32 in a sorted
         // set of 128 points). Similar to principle suggested by G4JNT.
         fltNoiseLevel = m_fltFftOutput[NBL_FFT_SIZE/8];
         // convert to decibels ...
         if(fltNoiseLevel>0)
                fltNoiseLevel = 20 * log10( fltNoiseLevel / 32767 );
          else  fltNoiseLevel = -200;
         // Normalize dB-SCALED result to a RX BANDwidth of 1 Hz,
         // to make it independent from the FFT BIN-width :
         // Example:
         //    A 'receiver' (here: a single FFT BIN..)
         //    with 0.5Hz bandwidth has a 'noise-reducing gain'
         //    of  +3.01 dB (= 10*log10(1.0Hz/0.5Hz) )
         //    compared to a receiver with 1.0 Hz bandwidth.
         //  SUBTRACT(!) the 'FFT gain' here, depending on the FFT bin width in Hertz.
         if( m_fltSampleRate>0 )
            fltNoiseLevel -= 10.0*log10( m_fltSampleRate / (T_Float)( 2 * NBL_FFT_SIZE ) );
         // TO TEST: use SpecLab's noise generator at 40 dB/sqrt(Hz) = max.
         m_fltCurrNoiseLevel = fltNoiseLevel;

         // Update the AVERAGE noise level which is important to detect sferics, etc
         m_fltAvrgNoisePower = 0.98 * m_fltAvrgNoisePower + 0.02 * fltNoiseLevel;

         // Update the PEAK noise level also.  Used for DISPLAY PURPOSES,
         // as a measure of "lightning activity", etc.
         if( fltNoiseLevel > m_fltAvrgNoisePower)
          {  m_fltPeakNoisePower = fltNoiseLevel;
          }
         else
          {  m_fltPeakNoisePower = 0.98 * m_fltPeakNoisePower + 0.02 * fltNoiseLevel;
          }

         // Calculate the new "blanking" multiplier.
         // > The noise contribution of each sample (ie. 10 ms block) to the FFT input
         // > would then have to be inversly proportional to its SNR. If the noise level
         // > WENT UP TEMPORARILY by 3 dB, we should reduce the gain by 6 dB.
         // Here, we are playing with LINEAR SCALED SAMPLES :
         m_fltCurrBlankingMultiplier = pow( 10,  // from DECIBEL to a FACTOR:
            -(m_fltPeakNoisePower - m_fltAvrgNoisePower)/(20/2) );
         // Avoid unnecessary "weak random amplitude modulation" (below 1dB)
         //  in the ABSENCE of pulse noise.
         if(  (m_fltCurrBlankingMultiplier > 0.891)
            &&(m_fltCurrBlankingMultiplier < 1.122) )
               m_fltCurrBlankingMultiplier = 1.0;
      } // end if <time to calculate the FFT>

#else  // no FFT ... but a delay line :


#endif // ( NOISEBLANKER_USE_FFT )

     // Much simpler, crude measure for the "average level",
     //  using a leaky integrator for the squared voltage
     //  (-> something proportional to the broadband signal power)
     fltNoisePower = fltTemp*fltTemp + 1e-10;
         // Note: Without adding 1e-10 here, the floating point unit gets
         //     VERY SLOW when processing "zero" input and the exponents
         //     of the numbers are "very negative" but the numbers are NON-ZERO.

     if( fltNoisePower > m_fltAvrgNoisePower * m_fltMinPeak2AvrgPowerRatio )
      { // momentary peak power level exceeds average level by programmed ratio ->
        //  Begin rising ramp, no matter what happened before !
        m_iRampDir = 1;  // ramp "up" (until total blanking sometime)
      } // end if < momentary peak power level exceeds average level by programmed ratio >
     else
      { // momentary peak power level does NOT exceed average level.
        // Begin decaying ramp only if the maximum is over !
        if(m_iRampDir==0)
           m_iRampDir=-1;
      }
     if(m_iRampDir != 0)
      { // something happening .. or happened not too long ago ?
        m_iRampIndex += m_iRampDir;
        if(m_iRampIndex>=m_iRampLength)
         { // reached the point of "maximum blanking" :
           m_iRampIndex=m_iRampLength-1;
           m_iRampDir = 0;  // may DECAY from now on, depending on trigger
         }
        else if(m_iRampIndex<=0)
         { // finished "ramp down". Stop blanking.
           m_iRampIndex = m_iRampDir = 0;
           m_fltCurrBlankingMultiplier = 1.0; // 'pass'
         }
        else
         { // ramp is somewhere between minimum and maximum now...
           if(m_iRampLength>0)
              m_fltCurrBlankingMultiplier = (T_Float)1.0 - (T_Float) m_iRampIndex/(T_Float) m_iRampLength;
           else
              m_fltCurrBlankingMultiplier = 0.0;
         }
      }

     // Lowpass filter for the "averaged" noise level.
     //   Must be slow enough not to react within a typical lightning "pulse train",
     //   otherwise the last pulse (of the train) may not be notched !
     //   With fs=44.1kHz  and NBL_AVRG_LP_STAGES=20,
     //     this lowpass "delays" the input by about 2 seconds,
     //     edge frequency (for -3dB) near 0.4 Hz, with reasonably steep transition.
     //     (found by trial-and-error)
     for(int i=0; i<NBL_AVRG_LP_STAGES; ++i)
      { fltNoisePower=m_fltAvLowpass[i]=0.9997*m_fltAvLowpass[i]+0.0003*fltNoisePower;
      }
     m_fltAvrgNoisePower = fltNoisePower;  // save for next chunk


     // Get oldest sample from buffer,
     //  multiply it with the current 'blanking multiplier',
     //  and place it in the caller's buffer :
     fltTemp = m_pFltBuffer[m_dwBufIndex];

#if(0) // TEST: lowpass for averaging..
     fltTemp = sqrt(m_fltAvrgNoisePower);
#endif // TEST


     *pDest++ = fltTemp * m_fltCurrBlankingMultiplier;

   } // end for( iSampleIndex=0; .. )

  return TRUE;
} // end CNoiseBlanker::ProcessRealSamples()


//***************************************************************************
T_Float CNoiseBlanker::GetNormalizedAvrgNoiseLevel(void)
{
  return m_fltAvrgNoisePower;
} // end CNoiseBlanker::GetNormalizedAvrgNoiseLevel()


//***************************************************************************
T_Float CNoiseBlanker::GetNormalizedPeakNoiseLevel(void)
{
  return 0;
} // end CNoiseBlanker::GetNormalizedPeakNoiseLevel()


// EOF < NoiseBlanker.cpp >
